// **************************************************************************
//
//    eses                   eses
//   eses                     eses
//  eses    eseses  esesese    eses   Embedded Systems Group
//  ese    ese  ese ese         ese
//  ese    eseseses eseseses    ese   Department of Computer Science
//  eses   eses          ese   eses
//   eses   eseses  eseseses  eses    University of Kaiserslautern
//    eses                   eses
//
// **************************************************************************
// This file implements an assembler for the instruction set architecture of
// the Abacus processor used for teaching and research by the Embedded
// Systems Group at the Univ. of Kaiserslautern.
//
// Instructions have 16 bits, whereas the uppermost 6 bits encode the opcode of 
// the instruction. The other bits depend on the opcode, where the following
// types of instructions can be distinguished:
//
//    R-Type (3 registers):
//        i{15:10} : opcode
//        i{9:7}   : index of destination register
//        i{6:4}   : index of register holding the left operand
//        i{3:1}   : index of register holding the right operand
//        i{0}     : additional R-type opcode, e.g. encoding vector/scalar op.
//    I-Type (2 registers; 1 immediate operand): 
//        i{15:10} : opcode
//        i{9:7}   : index of destination register
//        i{6:4}   : index of register holding the left operand
//        i{3:0}   : 4-bit constant operand contained in the instruction
//    S-Type (1 register; 1 immediate operand):
//        i{15:10} : opcode
//        i{9:7}   : index of destination register
//        i{6:0}   : 7-bit constant (either operand or function code)
//    J-Type (0 register; 1 immediate operand): 
//        i{15:10} : opcode
//        i{9:0}   : 10-bit constant for address offset
//
// This program converts assembler program files to Quartz programs. The 
// assembler programs may contain the following directives:
//  #DataWidth   8  : specifies the word size in bits
//  #MemSize    20  : memory size in words
//  #NumOfLines  4  : number of cache lines (if cache is used)
//  #LineSize    2  : size of one cache line (in words)
//  #SetAssoc    2  : set associativity of cache
//  #MissPenalty 4  : miss penalty for memory access
//  #memName        : name of memory module in directory memories
//  #cpuName        : name of processor module in directory cpus
//
// Single line comments start with "//", and lines holding an assembler
// instruction should have the following syntax:
//      label: instr p1,...,pn // comment
//
// The program implemented in this file then reads in such a file, then
// checks whether the sizes of the used immediate operands are small enough,
// and then converts the labels to integer distances. Finally, the
// instructions are encoded into 16-bit machine words, and a Quartz program
// is generated whose driver contains the encoded machine program.
// The modules listed under "memName" and "cpuName" are copied into the asm
// directory and the parameters of these files are adapted as specified in
// the assembler file. Simulation can thus be done by compiling and simulating
// the generated Quartz program.
// **************************************************************************

open System.IO

// ----------------------------------------------------------------------------
// exception with error message for reporting encountered problems
// ----------------------------------------------------------------------------

exception InternalException of string

let Error s = raise <| InternalException s

// ----------------------------------------------------------------------------
// parameters of the simulated computer system
// ----------------------------------------------------------------------------

let DataWidth   = ref  8
let MemSize     = ref 10
let NumOfLines  = ref  0
let LineSize    = ref  2
let SetAssoc    = ref  1
let MissPenalty = ref  0

let DirAbacus   = ref ""
let asmName     = ref ""
let cpuName     = ref "ScalarBehav"
let memName     = ref "MainMemory"


// ----------------------------------------------------------------------------
// replace macros in a Quartz file
// ----------------------------------------------------------------------------

let rec ReplaceMacros (sL:string list) =
    match sL with
    | [] -> []
    | s::sL ->
        let tL = Array.toList(s.Split([|' '|],System.StringSplitOptions.RemoveEmptyEntries))
        match tL with
        | "package"::tL -> ReplaceMacros sL
        | "import"::tL -> ReplaceMacros sL
        | "macro"::"DataWidth"::tL -> 
            let s = "macro DataWidth = "+(string(!DataWidth))+";"
            s::(ReplaceMacros sL)
        | "macro"::"MemSize"::tL -> 
            let s = "macro MemSize = "+(string(!MemSize))+";"
            s::(ReplaceMacros sL)
        | "macro"::"NumOfLines"::tL -> 
            let s = "macro NumOfLines = "+(string(!NumOfLines))+";"
            s::(ReplaceMacros sL)
        | "macro"::"LineSize"::tL -> 
            let s = "macro LineSize = "+(string(!LineSize))+";"
            s::(ReplaceMacros sL)
        | "macro"::"SetAssoc"::tL -> 
            let s = "macro SetAssoc = "+(string(!SetAssoc))+";"
            s::(ReplaceMacros sL)
        | "macro"::"MissPenalty"::tL -> 
            let s = "macro MissPenalty = "+(string(!MissPenalty))+";"
            s::(ReplaceMacros sL)
        | _ -> s::(ReplaceMacros sL)


// ----------------------------------------------------------------------------
// functions for reading and writing files
// ----------------------------------------------------------------------------

// read in text file line by line
let ReadLines (filename:string) =
    let stream = new StreamReader(filename)
    let rec ReadAllLines() = 
        let l = stream.ReadLine()
        if l=null then []
        else l::ReadAllLines()
    let L = ReadAllLines()
    stream.Close()
    L


// write string list as textfile
let WriteLines (filename:string) sL =
    let stream = new StreamWriter(filename)
    let rec WriteAllLines sL =
        match sL with
        | s::sL' ->
            let _ = stream.WriteLine(s:string)
            WriteAllLines sL'
        | [] -> ()
    WriteAllLines sL
    stream.Close()



// ----------------------------------------------------------------------------
// Function ParseLines below expects a list of strings and parses these into 
// a list of Abacus instructions. The result is a list of pairs (l,iL) where
// the the label l is of type string option and il is a list of strings 
// (machine instruction with its arguments). 
// For parsing, the following syntax is assumed:
//  * "//" denotes the beginning of a comment that ends with the line
//  * "#MemSize n"   defines the size of the main memory to words
//  * "#CacheSize n" defines the size of the data cache in words
//  * "#LineSize  n" defines the size of cache blocks in words
//  * "#SetAssoc n"  defines the set associativity of the cache
//  * lines holding an instruction must have the following form
//      label: instr p1,...,pn // comment
// ----------------------------------------------------------------------------

let ParseLines sL =
    // split a line into none-empty segments separated by characters in cL 
    let SplitNoneEmpty (cL: char []) (s:string) =
        let r = s.Split(cL,System.StringSplitOptions.RemoveEmptyEntries)
        Array.toList r
    // split a line into possibly empty segments separated by characters in cL 
    let SplitWithEmpty (cL: char []) (s:string) =
        let r = s.Split(cL,System.StringSplitOptions.None)
        Array.toList r
    // split a line by '/' to separate comments
    let splitComment s = SplitWithEmpty [|'/'|] s
    // remove comments of all lines sL
    let rec RemoveComments sL = 
        match sL with
        | [] -> []
        | l::sL' ->
            let tL = splitComment l
            if tL=[] then RemoveComments sL'
            elif List.head tL="" then RemoveComments sL'
            else (List.head tL)::(RemoveComments sL')
    // split a line by '#' to separate comments
    let setParameter defS =
        let tL = SplitNoneEmpty [|' '|] defS
        if List.length tL=2 then () 
        else Error("syntax error in parameter definition: "+defS)
        let paramS = List.nth tL 0
        let valueS = List.nth tL 1
        let str2int valueS = 
            try System.Int32.Parse valueS
            with _ -> Error("parameter "+valueS+" is not a number")
        match paramS with
        | "DataWidth"   -> DataWidth    := (str2int valueS)
        | "MemSize"     -> MemSize      := (str2int valueS)
        | "NumOfLines"  -> NumOfLines   := (str2int valueS)
        | "LineSize"    -> LineSize     := (str2int valueS)
        | "SetAssoc"    -> SetAssoc     := (str2int valueS)
        | "MissPenalty" -> MissPenalty  := (str2int valueS)
        | "cpuName"     -> cpuName      := valueS
        | "memName"     -> memName      := valueS
        | _ -> Error("parameter "+paramS+" not known")
    // the remaining lines are assumed to hold instructions which are either of
    // the form "label:instr" or "instr", so we split by ':', and after that by
    // ',' and ' ' to extract the parameters of CPU instructions
    let rec ParseLineL sL =
        match sL with
        | [] -> []
        | s::sL' ->
            let tL = SplitWithEmpty [|'#'|] s
            if List.length tL>1 then
                setParameter(List.head(List.rev tL))
                ParseLineL sL'
            else
                let tL = SplitNoneEmpty [|':'|] s
                let tL = List.map (SplitNoneEmpty [|' ';','|]) tL
                match tL with
                | [[ell];instrL] -> (Some(ell),instrL,s)::(ParseLineL sL')
                | [instrL]       -> (None,instrL,s)::(ParseLineL sL')
                | []   -> Error("syntax error in instruction "+s)
                | e::L -> Error("syntax error in instruction "+s)
    ParseLineL(RemoveComments sL)





// ----------------------------------------------------------------------------
// Functions MapLabels iL expects a list of Abacus instructions iL where jump
// and branch addresses are labels and replaces these labels with integers. The
// instruction list consists of pairs (l,i), where the label l is of type string
// option and i is a list of strings (machine instructions with its arguments).
// ----------------------------------------------------------------------------

let MapLabels iL = 
    // map labels to lines of corresponding instructions
    let rec mkLabelMap i iL =
        match iL with
        | (Some(ell),instL,instS)::iL' -> Map.add ell i (mkLabelMap (i+1) iL')
        | (None,instL,instS)::iL' -> mkLabelMap (i+1) iL'
        | [] -> Map.empty
    let labelMap = mkLabelMap 0 iL
    // replace labels by distances used in the machine instructions
    let rec unlabel i iL =
        match iL with
        | (_,[instr;r1;r2;ell],instS)::iL' ->
            if not(instr="bez" || instr="bnz") then 
                 ([instr;r1;r2;ell],instS)::(unlabel (i+1) iL')
            else
                let j = Map.tryFind ell labelMap
                if Option.isNone j then Error("label "+ell+" not declared")
                else
                    let c = (Option.get j) - i
                    ([instr;r1;r2;(string c)],instS)::(unlabel (i+1) iL')
        | (_,[instr;r1;ell],instS)::iL' ->
            if not(instr="jmp") then 
                 ([instr;r1;ell],instS)::(unlabel (i+1) iL')
            else
                let j = Map.tryFind ell labelMap
                if Option.isNone j then Error("label "+ell+" not declared")
                else
                    let c = (Option.get j) - i
                    ([instr;r1;(string c)],instS)::(unlabel (i+1) iL')
        | (_,["j";ell],instS)::iL' ->
            let j = Map.tryFind ell labelMap
            if Option.isNone j then Error("label "+ell+" not found")
            else
                let c = (Option.get j) - i
                (["j";(string c)],instS)::(unlabel (i+1) iL')
        | (_,instL,instS)::iL' -> (instL,instS)::(unlabel (i+1) iL')
        | [] -> []
    unlabel 0 iL





// ----------------------------------------------------------------------------
// encoding the instruction list and arguments in form of 16-bit opcodes
// ----------------------------------------------------------------------------

let Encode instrL =
    // convert n>=0 to radix-2 number
    let Nat2Radix2 n =
        let rec RevNat2BL n =
            match n with
            | 0 -> [false]
            | 1 -> [true]
            | _  -> ((n % 2)=1)::(RevNat2BL(n/2))
        if n<0 then Error("nat number should be non-negative")
        else List.rev(RevNat2BL n)
    // convert n to 2-complement number
    let Int2Compl2 n =
        let ceilDiv2 n = (n-1)/2 + 1
        let rec leqExp2 n = 
            match n with
            | 0 -> 1
            | 1 -> 1
            | 2 -> 2
            | n -> let e = leqExp2(ceilDiv2 n) in e+e
        if n>=0 then false::(Nat2Radix2 n)
        else
            let e = leqExp2(-n)
            let eL = Nat2Radix2 e
            let nL = Nat2Radix2(e+n)
            let rec ListRevDisj eL nL = 
                match eL,nL with
                | e::eL',n::nL' -> (e || n)::(ListRevDisj eL' nL')
                | _,[] -> eL
                | [],_ -> nL
            List.rev(ListRevDisj (List.rev eL) (List.rev nL))
    // print bool list to string
    let rec BL2Str bl =
        match bl with
        | [] -> ""
        | false::bl -> "0"+(BL2Str bl)
        | true::bl  -> "1"+(BL2Str bl)
    // replicate a bit for sign and zero extensions
    let rec replicate n b =
        if n=0 then []
        else b::(replicate (n-1) b)
    // zero extension of bl to l bits
    let ZE l ns =
        let n  = try System.Int32.Parse ns with _ -> Error(ns+" is not an integer")
        let bl = Nat2Radix2 n
        let len = List.length bl
        if l<len then Error("constant "+ns+" too large")
        else BL2Str((replicate (l-len) false)@bl)
    // sign extension of bl to l bits
    let SE l is =
        let i  = try System.Int32.Parse is with _ -> Error(is+" is not an integer")
        let bl = Int2Compl2 i
        let len = List.length bl
        if l<len then Error("constant "+is+" too large")
        else BL2Str((replicate (l-len) (List.head bl))@bl)
    // encode scalar registers
    let SR ri =
        match ri with
        | "R0" -> "000"
        | "R1" -> "001"
        | "R2" -> "010"
        | "R3" -> "011"
        | "R4" -> "100"
        | "R5" -> "101"
        | "R6" -> "110"
        | "R7" -> "111"
        | _ -> Error("unknown scalar register "+ri)
    // encode vector registers
    let VR vi =
        match vi with
        | "V0" -> "000"
        | "V1" -> "001"
        | "V2" -> "010"
        | "V3" -> "011"
        | "V4" -> "100"
        | "V5" -> "101"
        | "V6" -> "110"
        | "V7" -> "111"
        | _ -> Error("unknown vector register "+vi)
    // encode instruction 
    let encodeI instr =
        match instr with
        // scalar arithmetic R-type instructions
        | ["add";r1;r2;r3]   -> "000000"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["addu";r1;r2;r3]  -> "000001"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["sub";r1;r2;r3]   -> "000100"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["subu";r1;r2;r3]  -> "000101"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["mul";r1;r2;r3]   -> "001000"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["mulu";r1;r2;r3]  -> "001001"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["div";r1;r2;r3]   -> "001100"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["divu";r1;r2;r3]  -> "001101"+(SR r1)+(SR r2)+(SR r3)+"0"
        // scalar comparisons (R-type)
        | ["slt";r1;r2;r3]   -> "010000"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["sltu";r1;r2;r3]  -> "010001"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["sle";r1;r2;r3]   -> "010010"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["sleu";r1;r2;r3]  -> "010011"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["seq";r1;r2;r3]   -> "010100"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["sne";r1;r2;r3]   -> "010101"+(SR r1)+(SR r2)+(SR r3)+"0"
        // scalar logic instructions (R-type)
        | ["and";r1;r2;r3]   -> "010110"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["or";r1;r2;r3]    -> "010111"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["nand";r1;r2;r3]  -> "011000"+(SR r1)+(SR r2)+(SR r3)+"0"
        | ["nor";r1;r2;r3]   -> "011001"+(SR r1)+(SR r2)+(SR r3)+"0"
        // vector arithmetic R-type instructions
        | ["vadd";v1;v2;v3]  -> "000000"+(VR v1)+(VR v2)+(VR v3)+"1"
        | ["vaddu";v1;v2;v3] -> "000001"+(VR v1)+(VR v2)+(VR v3)+"1"
        | ["vsub";v1;v2;v3]  -> "000100"+(VR v1)+(VR v2)+(VR v3)+"1"
        | ["vsubu";v1;v2;v3] -> "000101"+(VR v1)+(VR v2)+(VR v3)+"1"
        | ["vmul";v1;v2;v3]  -> "001000"+(VR v1)+(VR v2)+(VR v3)+"1"
        | ["vmulu";v1;v2;v3] -> "001001"+(VR v1)+(VR v2)+(VR v3)+"1"
        | ["vdiv";v1;v2;v3]  -> "001100"+(VR v1)+(VR v2)+(VR v3)+"1"
        | ["vdivu";v1;v2;v3] -> "001101"+(VR v1)+(VR v2)+(VR v3)+"1"
        // vector comparisons (R-type)
        | ["vslt";r1;v2;v3]  -> "010000"+(SR r1)+(VR v2)+(VR v3)+"1"
        | ["vsltu";r1;v2;v3] -> "010001"+(SR r1)+(VR v2)+(VR v3)+"1"
        | ["vsle";r1;v2;v3]  -> "010010"+(SR r1)+(VR v2)+(VR v3)+"1"
        | ["vsleu";r1;v2;v3] -> "010011"+(SR r1)+(VR v2)+(VR v3)+"1"
        | ["vseq";r1;v2;v3]  -> "010100"+(SR r1)+(VR v2)+(VR v3)+"1"
        | ["vsne";r1;v2;v3]  -> "010101"+(SR r1)+(VR v2)+(VR v3)+"1"
        // vector logic instructions (R-type)
        | ["vand";v1;v2;v3]  -> "010110"+(VR v1)+(VR v2)+(VR v3)+"1"
        | ["vor";v1;v2;v3]   -> "010111"+(VR v1)+(VR v2)+(VR v3)+"1"
        | ["vnand";v1;v2;v3] -> "011000"+(VR v1)+(VR v2)+(VR v3)+"1"
        | ["vnor";v1;v2;v3]  -> "011001"+(VR v1)+(VR v2)+(VR v3)+"1"
        // scalar arithmetic I-type instructions
        | ["addi";r1;r2;c]   -> "000010"+(SR r1)+(SR r2)+(SE 4 c)
        | ["addiu";r1;r2;c]  -> "000011"+(SR r1)+(SR r2)+(ZE 4 c)
        | ["subi";r1;r2;c]   -> "000110"+(SR r1)+(SR r2)+(SE 4 c)
        | ["subiu";r1;r2;c]  -> "000111"+(SR r1)+(SR r2)+(ZE 4 c)
        | ["muli";r1;r2;c]   -> "001010"+(SR r1)+(SR r2)+(SE 4 c)
        | ["muliu";r1;r2;c]  -> "001011"+(SR r1)+(SR r2)+(ZE 4 c)
        | ["divi";r1;r2;c]   -> "001110"+(SR r1)+(SR r2)+(SE 4 c)
        | ["diviu";r1;r2;c]  -> "001111"+(SR r1)+(SR r2)+(ZE 4 c)
        // load and store instructions
        | ["ld";r1;r2;c]     -> "011010"+(SR r1)+(SR r2)+(ZE 4 c)
        | ["st";r1;r2;c]     -> "011011"+(SR r1)+(SR r2)+(ZE 4 c)
        | ["ll";r1;r2;c]     -> "011110"+(SR r1)+(SR r2)+(ZE 4 c)
        | ["sc";r1;r2;c]     -> "011111"+(SR r1)+(SR r2)+(ZE 4 c)
        | ["mov";r1;c]       -> "100000"+(SR r1)+(SE 7 c)
        | ["movu";r1;c]      -> "100001"+(SR r1)+(ZE 7 c)
        // branches and jumps
        | ["bez";r1;r2;c]    -> "100010"+(SR r1)+(SR r2)+(SE 4 c)
        | ["bnz";r1;r2;c]    -> "100011"+(SR r1)+(SR r2)+(SE 4 c)
        | ["jmp";r1;c]       -> "100100"+(SR r1)+(SE 7 c)
        | ["j";c]            -> "100101"+(SE 10 c)
        // dealing with cache, overflow register, vector mask and length registers
        | ["sync"]           -> "100111"+"000"+"0000000"
        | ["ovf"; r1]        -> "100111"+(SR r1)+"0000001"
        | ["mvtm";r1]        -> "100111"+(SR r1)+"0000010"
        | ["mvfm";r1]        -> "100111"+(SR r1)+"0000011"
        | ["mvtl";r1]        -> "100111"+(SR r1)+"0000100"
        | ["mvfl";r1]        -> "100111"+(SR r1)+"0000101"
        // pseudo instructions
        | ["nop"]            -> "000010"+"000"+"000"+"0000"          // addi(R0,R0,0)
        | ["neg";r1;r2]      -> "011001"+(SR r1)+(SR r2)+"000"+"0"   // nor(r1,r2,R0)
        | ["movr";r1;r2]     -> "000010"+(SR r1)+(SR r2)+"0000"      // addi(r1,r2,0)
        | ["sgt";r1;r2;r3]   -> "010000"+(SR r1)+(SR r3)+(SR r2)+"0" // slt(r1,r3,r2)
        | ["sgtu";r1;r2;r3]  -> "010001"+(SR r1)+(SR r3)+(SR r2)+"0" // sltu(r1,r3,r2)
        | ["sge";r1;r2;r3]   -> "010010"+(SR r1)+(SR r3)+(SR r2)+"0" // sle(r1,r3,r2)
        | ["sgeu";r1;r2;r3]  -> "010011"+(SR r1)+(SR r3)+(SR r2)+"0" // sleu(r1,r3,r2)
        | ["vsgt";r1;v2;v3]  -> "010000"+(SR r1)+(VR v3)+(VR v2)+"1" // vslt(r1,v3,v2)
        | ["vsgtu";r1;v2;v3] -> "010001"+(SR r1)+(VR v3)+(VR v2)+"1" // vsltu(r1,v3,v2)
        | ["vsge";r1;v2;v3]  -> "010010"+(SR r1)+(VR v3)+(VR v2)+"1" // vsle(r1,v3,v2)
        | ["vsgeu";r1;v2;v3] -> "010011"+(SR r1)+(VR v3)+(VR v2)+"1" // vsleu(r1,v3,v2)
        | ["jr";r1]          -> "100100"+(SR r1)+"0000000"           // jmp(r1,0);
        | iL -> Error("unknown instruction: "+(string iL))
    List.map (fun (instrL,instrS) -> ((encodeI instrL),instrS)) instrL




// ----------------------------------------------------------------------------
// writing the Quartz program of the parsed instructions
// ----------------------------------------------------------------------------

let WriteQuartz opcL =
    let MemSizeS     = string(!MemSize)
    let NumOfLinesS  = string(!NumOfLines)
    let LineSize S   = string(!LineSize )
    let SetAssocS    = string(!SetAssoc)
    let MissPenaltyS = string(!MissPenalty)
    let opcLlen = List.length opcL
    let stream = new StreamWriter((!asmName)+".qrz")
    let writeln s = stream.WriteLine(s:string)
    let rec WriteAllLines sL =
        match sL with
        | s::sL' ->
            writeln(s:string)
            WriteAllLines sL'
        | [] -> ()
    let rec WriteAllCmds i sL =
        match sL with
        | (opc,s)::sL' ->
            let is = sprintf "%3d: " i
            let s' = if sL'=[] then "   // "+is else ",  // "+is
            let _ = writeln(("        0b"+opc+s'+s):string)
            WriteAllCmds (i+1) sL'
        | [] -> ()
    writeln("// ************************************************************************** //")
    writeln("//                                                                            //")
    writeln("//    eses                   eses                                             //")
    writeln("//   eses                     eses                                            //")
    writeln("//  eses    eseses  esesese    eses   Embedded Systems Group                  //")
    writeln("//  ese    ese  ese ese         ese                                           //")
    writeln("//  ese    eseseses eseseses    ese   Department of Computer Science          //")
    writeln("//  eses   eses          ese   eses                                           //")
    writeln("//   eses   eseses  eseseses  eses    University of Kaiserslautern            //")
    writeln("//    eses                   eses                                             //")
    writeln("//                                                                            //")
    writeln("// ************************************************************************** //")
    writeln("")
    writeln("                                                                           ")
    writeln("macro DataWidth = "+(string(!DataWidth))+";      // bit-width of registers ")
    writeln("                                                                           ")
    writeln("module "+(!asmName)+"(bv{16} ?instr,nat pc) {                              ")
    writeln("    event nat adrMem;                 // address for memory access         ")
    writeln("    event bv{DataWidth} dataMem;      // data for memory access            ")
    writeln("    event readMem,writeMem;           // flags for reading/writing         ")
    writeln("    event reqMem,ackMem,doneMem;      // signals for memory transaction    ")
    writeln("    // single CPU")
    writeln("    cpu: "+(!cpuName)+"(instr,pc,adrMem,dataMem,readMem,writeMem,reqMem,ackMem,doneMem);")
    writeln("    ||")
    writeln("    // connected to data memory")
    writeln("    memory: "+(!memName)+"(adrMem,dataMem,readMem,writeMem,reqMem,ackMem,doneMem);")
    writeln("}")
    writeln("drivenby "+(!asmName)+"Drv {")
    writeln("    ["+(string opcLlen)+"]bv{16} Prog;")
    writeln("    Prog = [")
    WriteAllCmds 0 opcL
    writeln("    ];")
    writeln("    pause;")
    writeln("    weak abort {")
    writeln("        loop {")
    writeln("            instr = Prog[pc];")
    writeln("            pause;")
    writeln("        }")
    writeln("    } when(pc>="+(string (opcLlen-1))+");")
    writeln("}")
    stream.Close()


// ----------------------------------------------------------------------------
// get parameters of the command line
// ----------------------------------------------------------------------------

let rec getParameters argL =
    match argL with
    | []  -> Error("usage: asm2qrz -cpu <cpu_file> -mem <mem_file> <asm_file")
    | [s] -> s
    | "-cpu"::v::argL -> 
        cpuName := v
        getParameters argL
    | "-mem"::v::argL -> 
        memName := v
        getParameters argL
    | _ -> Error("usage: asm2qrz -cpu <cpu_file> -mem <mem_file> <asm_file")


// ----------------------------------------------------------------------------
// modify parameters in a Quartz file
// ----------------------------------------------------------------------------

let WriteParameterizedFile indir outdir inFile =
    let inFilename  = Path.Combine(!DirAbacus,Path.Combine(indir, inFile+".qrz"))
    let outFilename = Path.Combine(!DirAbacus,Path.Combine(outdir,inFile+".qrz"))
    let sL = ReadLines inFilename
    WriteLines outFilename (ReplaceMacros sL)


// ----------------------------------------------------------------------------
// the main program
// ----------------------------------------------------------------------------

[<EntryPoint>]
let main (args : string []) =
    try
        // determine parameters
        let argL = Array.toList args
        let asmFile = getParameters argL
        asmName := Path.GetFileNameWithoutExtension asmFile
        // determine path to Abacus directory
        let curr = Directory.GetCurrentDirectory()
        DirAbacus := Path.GetDirectoryName curr
        // parse asm file
        let sL = ReadLines asmFile
        let iL = ParseLines sL
        // replace labels
        let inL = MapLabels iL
        // encode the instructions
        let opcL = Encode inL
        WriteQuartz opcL
        // read cpu and memory files
        let _ = WriteParameterizedFile "cpus" "asm" (!cpuName)
        let _ = WriteParameterizedFile "memories" "asm" (!memName)
        0
    with InternalException(s) ->
        printf "error: %s\n" s
        1
